@@ -56,7 +56,7 @@ static PyObject *Error, *Warning, *InterfaceError,
 	*DatabaseError, *InternalError, *OperationalError, *ProgrammingError,
 	*IntegrityError, *DataError, *NotSupportedError;
 
-static const char *PyPgVersion = "0.9.1";
+static const char *PyPgVersion = "0.9.2";
 static char pg__doc__[] = "Simple Python interface to PostgreSQL DB";
 
 /* default values */
@@ -528,7 +528,19 @@ static pgparams *_pgsource_getparams(PyObject *params)
 	    ret->paramTypes[i] = BOOLOID;
 	} else if (PyInt_Check(param) || PyLong_Check(param)) {
 	    PyObject *str = PyObject_Str(param);
-	    ret->paramValues[i] = PyString_AsString(str);
+	    ret->paramValues[i] = strdup(PyString_AsString(str));
+	    if (ret->paramValues[i] == NULL) {
+		Py_DECREF(param);
+		Py_DECREF(str);
+		_pgsource_freeparams(ret);
+		PyErr_SetString(ProgrammingError, "out of memory binding paramaters");
+		return NULL;
+	    }
+	    ret->mustFree[i] = 1;
+	    if (PyInt_Check(param))
+		ret->paramTypes[i] = INT4OID;
+	    else
+		ret->paramTypes[i] = INT8OID;
 	    Py_DECREF(str);
 	} else if (PyString_Check(param)) {
 	    int len;
@@ -551,6 +563,12 @@ static pgparams *_pgsource_getparams(PyObject *params)
 	    /* this is kind of lame - could not find any documentation
 	       how to pass a double as a binary */
 	    char *dblstr = malloc(128);
+	    if (dblstr == NULL) {
+		Py_DECREF(param);
+		_pgsource_freeparams(ret);
+		PyErr_SetString(ProgrammingError, "out of memory binding paramaters");
+		return NULL;
+	    }
 	    snprintf(dblstr, 128, "%f", PyFloat_AsDouble(param));
 	    ret->paramValues[i] = dblstr;
 	    ret->mustFree[i] = 1;
@@ -568,8 +586,19 @@ static pgparams *_pgsource_getparams(PyObject *params)
 
 	    if (o != NULL && PyCallable_Check(o)) {
 		int len;
+		char *value;
 		str = PyObject_CallObject(o, NULL);
-		PyString_AsStringAndSize(str, &(ret->paramValues[i]), &len);
+		PyString_AsStringAndSize(str, &value, &len);
+		ret->paramValues[i] = calloc(1, len+1);
+		if (ret->paramValues[i] == NULL) {
+		    Py_DECREF(param);
+		    Py_DECREF(str);
+		    _pgsource_freeparams(ret);
+		    PyErr_SetString(ProgrammingError, "out of memory binding paramaters");
+		    return NULL;
+		}
+		memcpy(ret->paramValues[i], value, len);
+		ret->mustFree[i] = 1;
 		ret->paramTypes[i] = TEXTOID;
 		ret->paramLengths[i] = len;
 		Py_DECREF(str);
@@ -1083,13 +1112,17 @@ pgsource_fetchall(pgsourceobject * self, PyObject * args)
     if (!check_no_args(args, "fetchall"))
 	return NULL;
 
-    /* should we return anything? */
-    if (self->current_row >= self->max_row) {
+    /* need to return an empty list ? */
+    if (self->current_row > self->max_row) {
+	/* already returned the empty list, another fetchall call is invalid... */
 	Py_INCREF(Py_None);
 	return Py_None;
+    } else if (self->current_row == self->max_row) {
+	self->current_row++;
+	return PyList_New(0);
     }
 
-    /* stores result in tuple */
+    /* stores result in presized list for efficiency */
     reslist = PyList_New(self->max_row);
 
     /* return the remaining rows that have not been "extracted" yet */
@@ -1102,7 +1135,7 @@ pgsource_fetchall(pgsourceobject * self, PyObject * args)
 	PyList_SET_ITEM(reslist, row, rowtuple);
     }
     /* mark all rows returned */
-    self->current_row = self->max_row;
+    self->current_row = self->max_row + 1;
     /* returns list */
     return reslist;
 }
@@ -1122,12 +1155,17 @@ pgsource_fetchalldict(pgsourceobject * self, PyObject * args)
     if (!check_no_args(args, "fetchall"))
 	return NULL;
 
-    /* should we return anything? */
-    if (self->current_row >= self->max_row) {
+    /* need to return an empty list ? */
+    if (self->current_row > self->max_row) {
+	/* already returned the empty list, another fetchall call is invalid... */
 	Py_INCREF(Py_None);
 	return Py_None;
+    } else if (self->current_row == self->max_row) {
+	self->current_row++;
+	return PyList_New(0);
     }
-    /* stores result in tuple */
+
+    /* stores result in presized list for efficiency */
     reslist = PyList_New(self->max_row);
 
     /* return the remaining rows that have not been "extracted" yet */
@@ -1140,7 +1178,7 @@ pgsource_fetchalldict(pgsourceobject * self, PyObject * args)
 	PyList_SET_ITEM(reslist, row, rowdict);
     }
     /* mark all rows returned */
-    self->current_row = self->max_row;
+    self->current_row = self->max_row + 1;
     /* returns list */
     return reslist;
 }
